#include <ESPmDNS.h>


#include "esp_camera.h"
#include <TMCStepper.h>

#include <WiFi.h>
#include <ESPAsyncWebServer.h>
#include <ArduinoJson.h>

#include "esp_timer.h"
#include "img_converters.h"

#include <SPIFFS.h>

#include "soc/soc.h"           // Disable brownout problems
#include "soc/rtc_cntl_reg.h"  // Disable brownout problems
#include <ESPmDNS.h>
#include <Update.h>

// Yes ESP32cam

#define U_PART U_SPIFFS

const char *SSID = "NetworkCongestion";
const char *PWD = "mp@ssword";
const char *hostname = "barndoor";
const char *version = "1.4";

const char *FILE_PHOTO = "/photo.jpg";

#define SERIAL_PORT Serial2 // HardwareSerial port 
#define DRIVER_ADDRESS 0b00
#define ENABLE_PIN 14
#define RED_LED 33
#define FLASH_PIN 4

// #define TOUCH_PIN 12

#define DIAG_PIN 13
#define INDEX_PIN 12

#define STALL_VALUE  10 // [-64..63]

#define RXD2 15
#define TXD2 2

// using namespace TMC2209_n;

#define R_SENSE 0.11f // Match to your driver
                      // SilentStepStick series use 0.11
                      // UltiMachine Einsy and Archim2 boards use 0.2
                      // Panucatt BSD2660 uses 0.1
                      // Watterott TMC5160 uses 0.075


// OV2640 camera module pins (CAMERA_MODEL_AI_THINKER)
#define PWDN_GPIO_NUM     32
#define RESET_GPIO_NUM    -1
#define XCLK_GPIO_NUM      0
#define SIOD_GPIO_NUM     26
#define SIOC_GPIO_NUM     27

#define Y9_GPIO_NUM       35
#define Y8_GPIO_NUM       34
#define Y7_GPIO_NUM       39
#define Y6_GPIO_NUM       36
#define Y5_GPIO_NUM       21
#define Y4_GPIO_NUM       19
#define Y3_GPIO_NUM       18
#define Y2_GPIO_NUM        5
#define VSYNC_GPIO_NUM    25
#define HREF_GPIO_NUM     23
#define PCLK_GPIO_NUM     22

TMC2209Stepper driver(&SERIAL_PORT, R_SENSE, DRIVER_ADDRESS);

AsyncWebServer server(80);
AsyncEventSource events("/events");

int accel = 0;
int stall = STALL_VALUE;
bool shaft = false;
volatile int direction = 0;

bool hasstorage = false;
bool hascamera = false;
bool islistening = false;
volatile bool autocloseenabled = true;

volatile bool autoclose = false;
volatile unsigned long autoclose_timer = 0;
volatile unsigned long autoclose_delay = 20000;

volatile bool takepicture = false;
volatile bool get_motor_status = false;

unsigned long opentiming = 14500;
unsigned long closetiming = 14500;
unsigned long stepstoopen = 1385;
unsigned long stepstoclose = 1385;

unsigned long timetoopen = 14500;
unsigned long timetoclose = 14500;

volatile bool bumped = false;
// volatile bool touch1detected = false;

volatile long steps;


TaskHandle_t Motor;

void IRAM_ATTR bump(void* arg)
{
    bumped = true;
}

void IRAM_ATTR step(void* arg)
{
    steps = steps +1;
}

// void gotTouch1(){
//  touch1detected = true;
// }

StaticJsonDocument<5*1024> jsonDocument;
char buffer[1024];

class LogClass : public Print
{
  private:
  unsigned long line, len;
  char logbuffer[2048] = "";
  char temp[1024];
  char *a, *b, *c;
  
  char buff[4096];

  public:

    virtual size_t write(uint8_t c) {
      Serial.write(c);

      *temp = c;
      *(temp+1) = '\0';
      
      if (islistening) events.send(temp,"logger",millis());

      addtobuffer();

      return 1;
    }

    virtual size_t write(const uint8_t *s, size_t size)
    {
      Serial.write(s, size);
      
      for (int i=0; i < size; i++)
      { 
          *(a +i) = *(s + i);
      }
      *(temp+size) = '\0';

      if (islistening) events.send(temp,"logger",millis());
      
      addtobuffer();

      return 1;
    }

    size_t write(const char *buffer, size_t size)
    {
      return write((const uint8_t *) buffer, size);
    }

    LogClass()
    {
      unsigned long tea = millis();

      Serial.begin(115200);
      while (Serial.available()==0 && ((millis() - tea) < 4000));

      a = temp;
      b = logbuffer;
      line = 1;
      len = 0;
    }

    void addtobuffer()
    {
      unsigned long i, l1 = len, l2 = 0;
      
      c = a;
      
      while(*(c + l2) != '\0') l2++; //strlen(a)
        
      while ((l1 + l2) > 2048) //pop oldest off
      {
        c = b;
        i = 0;
        while (*(c + i) != '\n') i++; //find end of first line
        c += i + 1;
        i = 0;

        while (*(c + i) != '\0') { //shift the rest 
          *(b + i) = *(c + i);
          i++;
        }
        *(b + i) = '\0';
        l1 = i;
        line++;
      }

      len = l1;
      
      i = 0;
      c = a;

      while(*(c + i) != '\0') // append
      {
        *(b + len + i) = *(c + i);
        i++;
      }
      len += i;

      *(b + len) = '\0';
    }

    char *Buffer()
    {
      snprintf(buff, 4095, "<HTML><HEAD></HEAD><BODY><div>Log Contents %d</div><pre>%s</pre></BODY>", strlen(b), b);
      return buff;
    }

    char *JSON()
    {
      int i = 0, l = line;
      char t;
      JsonObject obj;

      jsonDocument.clear();

      c = b;
      while (*c != '\0')
      {
        i = 0;
        while (*(c + i) != '\n') 
        {
          *(a + i) = *(c + i);
          i++;
        }
        *(a + i) = '\0';

        obj = jsonDocument.createNestedObject();
        obj["type"] = l;
        obj["value"] = a;

        c += i + 1;
        l++;
      }

      serializeJson(jsonDocument, buff); 
     // Serial.println(buff);
      return buff;
    }
} Log;


// web page

const char index_html[] PROGMEM = R"rawliteral(
<!DOCTYPE HTML><html>
<head>
  <meta name="viewport" content="width=device-width, initial-scale=1">
  <style>
    body { text-align:center; }
    .vert { margin-bottom: 1%; }
    .hori{ margin-bottom: 0%; }
    .center {
       margin-left: auto;
       margin-right: auto;
    }
    .parent {
      margin: 1rem;
      padding: 0rem 0rem;
      text-align: center;
      vertical-align: top;
    }
    .child {
      display: inline-block;
      padding: 1rem 1rem;
      text-align: left;
      vertical-align: top;
    }
  </style>
</head>
<body>
  <div class='parent' id="container">
    <div class='child'>
      <form id='parameters'>
        <label for="opentiming">Open Timing (ms):</label><br>
        <input type="number" value="14500" id="opentiming" name="opentiming"><br>
        
        <label for="closetiming">Close Timing (ms):</label><br>
        <input type="number" value="14500" id="closetiming" name="closetiming"><BR>

         <label for="stepstoopen">Steps to open:</label><br>
        <input type="number" value="1385" id="stepstoopen" name="stepstoopen"><br>
        
        <label for="Stepstoclose">steps to close:</label><br>
        <input type="number" value="1385" id="stepstoclose" name="stepstoclose"><BR>

        <label for="autoclosedelay">AutoClose Delay (ms):</label><br>
        <input type="number" value="20000" id="autoclosedelay" name="autoclosedelay"><br>

        <p><input type="checkbox" value="1" id="autocloseenabled" name="autocloseenabled" Checked>
          <label for="autocloseenabled">AutoClose Enabled</label></p>
      </form>
    </div>
    <div class='child'>
      <h2>the Lab</h2>
      <p>
        <button onclick="OpenDoor();">Open Door</button>
        <button onclick="CloseDoor()">Close Door</button>
        <button onClick="ViewLog()">View Log</button>
        <button onClick="UpdateFirmware()">Update Firmware</button>
        <button onClick="Reboot()">Reboot</button>
      </p>
    </div>
    <div id="stats" class='child'>.</div>
  </div>
  <div class='parent'><img onclick="TakePic();" id="photo" src="saved-photo" width="30%" class='child'></img><div id = "log" style="overflow: auto; max-height: 30vh;" class='child'>.</div></div>
</body>
<script>

  if (!!window.EventSource) {
    var source = new EventSource('/events');

    source.addEventListener('open', function(e) {
      console.log("listening to door");
    }, false);

    source.addEventListener('error', function(e) {
      if (e.target.readyState != EventSource.OPEN) {
        console.log("door is silent");
      }
    }, false);

    source.addEventListener('motor', function(e) {
      console.log("motor", e.data);
      var myArr = JSON.parse(e.data);
      FillTable(myArr, "stats");
    }, false);
    
    source.addEventListener('logger', function(e) {
      console.log("Log", e.data);
    }, false);
    
    source.addEventListener('progress', function(e) {
      document.getElementById("stats").innerHTML = '<div style="width:200px;background-color:#808080"><div style="background-color:#ABBAEA; text-align:center; height:24px; width:' + e.data + '">' + e.data + '</div></div>';
    }, false);

    source.addEventListener('camera', function(e) {
      var timestamp = new Date().getTime(); 
      console.log("camera", e.data);
      document.getElementById("photo").src="saved-photo?t=" + timestamp;
    }, false);
  }

  function FillTable(arr, id) {
    var out = "<table style='max-width:300px;'>";
    var i;
    for(i = 0; i < arr.length; i++) {
        out += "<tr><td style='text-align:right;width:5%'>" + arr[i].type + "</td><td style='white-space:nowrap;width:95%;text-align:left'>" + arr[i].value + '</td>' + '</tr>';
    }
    out += "</table";
    document.getElementById(id).innerHTML = out;
  }

  function OpenDoor() {
    var xhr = new XMLHttpRequest();
    xhr.open('POST', "/Open");

    xhr.onload = function(event){ 
      var myArr = JSON.parse(event.target.response);
      FillTable(myArr, "stats");
    }; 

    var formData = new FormData(document.getElementById("parameters")); 
    var s = JSON.stringify(Object.fromEntries( formData.entries()));
    xhr.send(s);
  }

  function CloseDoor() {
    var xhr = new XMLHttpRequest();
    xhr.open('POST', "/Close");
    
    xhr.onload = function(event){ 
      var myArr = JSON.parse(event.target.response);
      FillTable(myArr, "stats");
    }; 

    var formData = new FormData(document.getElementById("parameters")); 
    var s = JSON.stringify(Object.fromEntries( formData.entries()));
    xhr.send(s);
  }
  
  function Reboot() {
    var xhr = new XMLHttpRequest();
    xhr.open('GET', "/Reboot", true);
    xhr.send();
   }

  function TakePic(){
    var xhr = new XMLHttpRequest();
    xhr.open('POST', "/pic");
    
    xhr.onload = function(event){ 
      var myArr = JSON.parse(event.target.response);
      FillTable(myArr, "stats");
    }; 
    
    var formData = new FormData(document.getElementById("parameters")); 
    xhr.send(JSON.stringify(Object.fromEntries( formData.entries())));
  }

  function ViewLog() {
    xhr = new XMLHttpRequest();
    xhr.onreadystatechange = function() {
        if (this.readyState == 4 && this.status == 200) {
            var myArr = JSON.parse(this.responseText);
            FillTable(myArr, "log");
        }
    };
    xhr.open('GET', "/log", true);
    xhr.send(); 
  }

  function UpdateFirmware(){
    document.getElementById("stats").innerHTML ="<form method='POST' action='/doUpdate' enctype='multipart/form-data'><input type='file' name='update'><input type='submit' value='Update'></form>"
  }
</script>
</html>)rawliteral";


void setupMDNS()
{
  /*use mdns for host name resolution*/
  if (!MDNS.begin(hostname)) { //http://esp32.local
    Log.printf("Error setting up MDNS responder!\n");
    while (1) {
      delay(1000);
    }
  }
  MDNS.addService("http", "tcp", 80);
  Log.printf("mDNS responder started\n");
}

void Mount_Storage()
{
  if (!SPIFFS.begin(true)) {
    Log.printf("An Error has occurred while mounting SPIFFS\n");
    ESP.restart();
  }
  else {
    delay(500);
    Log.printf("SPIFFS mounted successfully\n");
    SPIFFS.format();
    hasstorage = true;
  }
}

void connectToWiFi() {
  bool flip = false;
  Log.printf("Connecting to %s...", SSID);
  
  WiFi.mode(WIFI_STA);
  WiFi.config(INADDR_NONE, INADDR_NONE, INADDR_NONE, INADDR_NONE);
  WiFi.setHostname(hostname); //define hostname

  WiFi.begin(SSID, PWD);
  
  while (WiFi.status() != WL_CONNECTED) {
    digitalWrite(RED_LED, flip ? HIGH : LOW);
    delay(500);
    flip = !flip;
    Log.print(".");
  }
  digitalWrite(RED_LED, HIGH);

  Log.printf("Connected. IP: %d:%d:%d:%d\nRRSI: %d\n", WiFi.localIP()[0], WiFi.localIP()[1], WiFi.localIP()[2], WiFi.localIP()[3], WiFi.RSSI());  //Just go with it
 
  WRITE_PERI_REG(RTC_CNTL_BROWN_OUT_REG, 0);

}

void apply_json_settings(uint8_t *data, size_t len)
{
    DynamicJsonDocument doc(2048);
    unsigned long h, i, j, k, l, m;

    // *(data + len) = '\0';
    // Log.printf("%s\n", data);

    deserializeJson(doc, data);
    i = doc["opentiming"];
    j = doc["closetiming"];
    k = doc["autoclosedelay"];
    h = doc["autocloseenabled"] == "1";
    l = doc["stepstoopen"];
    m = doc["stepstoclose"];
    Log.printf("opentiming: %d, closetiming: %d\n", i, j);
    Log.printf("stepstoopen:%d, stepstoclose:%d\n",l, m);
    Log.printf("autocloseenabled %s, autoclosedelay: %d\n",  h ? "yes" : "no", k);
    opentiming = i;
    closetiming = j;
    autoclose_delay = k;
    autocloseenabled = h;
    stepstoopen = l;
    stepstoclose = m;
}

 
void add_json_object(char *tag, float value) {
  JsonObject obj = jsonDocument.createNestedObject();
  obj["type"] = tag;
  obj["value"] = value;
}

char *get_stats()
{
  jsonDocument.clear();
  add_json_object("Bumped", bumped);
  add_json_object("Steps", steps);
  add_json_object("TimeToOpen", timetoopen);
  add_json_object("TimeToClose", timetoclose);
  serializeJson(jsonDocument, buffer);
  return buffer;
}

void OpenDoor(AsyncWebServerRequest * request)
{
  direction = -1;
  get_motor_status = true;
  autoclose = autocloseenabled;
  request->send(200, "application/json", get_stats());
}

void CloseDoor(AsyncWebServerRequest * request)
{
  direction = 1;
  get_motor_status = true;
  request->send(200, "application/json", get_stats());
}

void TakePic(AsyncWebServerRequest * request){
  takepicture = true;

  AsyncWebServerResponse *response = request->beginResponse(200,  "application/json", get_stats()); 
  response->addHeader("Refresh", "10");  
  response->addHeader("Location", "/");
  request->send(response);  
}

// setup API resources
void setup_routing() {
 
  server.on("/Open", HTTP_POST, [](AsyncWebServerRequest * request){}, NULL, [](AsyncWebServerRequest * request , uint8_t *data, size_t len, size_t index, size_t total) {
    apply_json_settings(data, len);
    OpenDoor(request);
  });

  server.on("/Close", HTTP_POST, [](AsyncWebServerRequest * request){}, NULL, [](AsyncWebServerRequest * request , uint8_t *data, size_t len, size_t index, size_t total) {
    apply_json_settings(data, len);
    CloseDoor(request);
  });

  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send_P(200, "text/html", index_html);
  });

  server.on("/pic", HTTP_POST, [](AsyncWebServerRequest *request){
    TakePic(request);
  });

  server.on("/saved-photo", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(SPIFFS, FILE_PHOTO, "image/png");
  });
  
  server.on("/update", HTTP_GET, [](AsyncWebServerRequest *request){handleUpdate(request);});

  server.on("/doUpdate", HTTP_POST,
    [](AsyncWebServerRequest *request) {},
    [](AsyncWebServerRequest *request, const String& filename, size_t index, uint8_t *data,
                  size_t len, bool final) {handleDoUpdate(request, filename, index, data, len, final);}
  );

  server.on("/logpage", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(200, "text/html", Log.Buffer());
    });
  
  server.on("/log", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(200, "application/json", Log.JSON());
    });

  server.on("/version", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send_P(200, "text/text", version);
  });

  server.on("/Reboot", HTTP_GET, [](AsyncWebServerRequest *request){
    AsyncWebServerResponse *response = request->beginResponse(302, "text/plain", "Please wait while the device reboots");
    response->addHeader("Refresh", "20");  
    response->addHeader("Location", "/");
    request->send(response);
    ESP.restart();
  });

  events.onConnect([](AsyncEventSourceClient *client){
    if(client->lastId()){
      Log.printf("Client reconnected! Last message ID that it got is: %u\n", client->lastId());
    }
    client->send("Friends, Romans, countrymen..",NULL,millis(),1000);
  });

  //HTTP Basic authentication
  // events.setAuthentication("user", "pass");

  server.addHandler(&events);

  // start server
  server.begin();
  islistening = true;
}

void setup_TMC2209()
{
  digitalWrite(ENABLE_PIN ,LOW);

  pinMode(DIAG_PIN ,INPUT);
  pinMode(INDEX_PIN, INPUT); //How far did it go?

  digitalWrite(RED_LED, HIGH); // off

  Serial2.begin(115200, SERIAL_8N1, RXD2, TXD2);     // INITIALIZE UART to TMC2208
  
  driver.begin();                // Initialize driver
                           
  driver.toff(5);                 // Enables driver in software

  driver.rms_current(800);       // Set motor RMS current
  Log.printf("current set to: %d\n", driver.rms_current());
 
  driver.microsteps(256); 
           
  driver.TCOOLTHRS(0xFFFFF); // 20bit max
  driver.TPWMTHRS(0); 
  driver.semin(5);
  driver.semax(2);
  driver.sedn(0b01);
  driver.SGTHRS(STALL_VALUE);
  driver.pwm_autoscale(true);   // Needed for stealthChop
  driver.en_spreadCycle(false);   // Toggle spreadCycle for smooth & silent operation
  driver.blank_time(24);
  driver.iholddelay(3); 

  //attachInterrupt(digitalPinToInterrupt(DIAG_PIN), bump, RISING);
  //attachInterrupt(digitalPinToInterrupt(INDEX_PIN), step, FALLING);

  //Some crap about the camera means more arcane looking code.
  gpio_install_isr_service(0);
  gpio_set_intr_type(GPIO_NUM_13, GPIO_INTR_POSEDGE); //DIAG_PIN
  gpio_isr_handler_add(GPIO_NUM_13, bump, NULL);

  gpio_set_intr_type(GPIO_NUM_12, GPIO_INTR_POSEDGE); //INDEX_PIN
  gpio_isr_handler_add(GPIO_NUM_12, step, NULL);
}

void setup_camera()
{
    // OV2640 camera module
  camera_config_t config;
  config.ledc_channel = LEDC_CHANNEL_0;
  config.ledc_timer = LEDC_TIMER_0;
  config.pin_d0 = Y2_GPIO_NUM;
  config.pin_d1 = Y3_GPIO_NUM;
  config.pin_d2 = Y4_GPIO_NUM;
  config.pin_d3 = Y5_GPIO_NUM;
  config.pin_d4 = Y6_GPIO_NUM;
  config.pin_d5 = Y7_GPIO_NUM;
  config.pin_d6 = Y8_GPIO_NUM;
  config.pin_d7 = Y9_GPIO_NUM;
  config.pin_xclk = XCLK_GPIO_NUM;
  config.pin_pclk = PCLK_GPIO_NUM;
  config.pin_vsync = VSYNC_GPIO_NUM;
  config.pin_href = HREF_GPIO_NUM;
  config.pin_sscb_sda = SIOD_GPIO_NUM;
  config.pin_sscb_scl = SIOC_GPIO_NUM;
  config.pin_pwdn = PWDN_GPIO_NUM;
  config.pin_reset = RESET_GPIO_NUM;
  config.xclk_freq_hz = 20000000;
  config.pixel_format = PIXFORMAT_JPEG;

  if (psramFound()) {
    config.frame_size = FRAMESIZE_UXGA;
    config.jpeg_quality = 10;
    config.fb_count = 2;
  } else {
    config.frame_size = FRAMESIZE_SVGA;
    config.jpeg_quality = 12;
    config.fb_count = 1;
  }
  // Camera init
  esp_err_t err = esp_camera_init(&config);

  if (err != ESP_OK) {
    Log.printf("Camera init failed with error 0x%x\n", err);
    // ESP.restart();
    hascamera = false;
  }
  else
    hascamera = true;
    
}

void setup() {

  // touchAttachInterrupt(TOUCH_PIN, gotTouch1, 40);
  pinMode(ENABLE_PIN ,OUTPUT);
  digitalWrite(ENABLE_PIN ,HIGH);
  pinMode(FLASH_PIN ,OUTPUT);
  pinMode(RED_LED, OUTPUT);
  
  delay(1000);
  setup_camera();
  Log.printf("camera setup %s.\n", (hascamera ? "OK": "Failed"));
  Mount_Storage();
  Log.printf("Storage mounted.\n");

  if (hascamera) {
    digitalWrite(FLASH_PIN ,HIGH);
    capturePhotoSaveSpiffs();
  }
  steps = 0;

  setup_TMC2209();
  Log.printf("Motor setup.\n");
  connectToWiFi();
  Log.printf("WIFI setup.\n");
  setup_routing(); 
  Log.printf("Routing setup.\n");
  setupMDNS();
  xTaskCreatePinnedToCore(MoveDoor, "Motor", 5000,NULL,5,&Motor,1);
}

void loop() {
    if (hascamera && hasstorage && takepicture)
    {
      takepicture = false;
      hascamera = false;
      digitalWrite(FLASH_PIN ,HIGH);
      vTaskDelay(500);
      capturePhotoSaveSpiffs();
      events.send("new picture available","camera",millis());
      hascamera = true;  
    }

    vTaskDelay(1000);
 }

unsigned long move(int start, int spd, bool dir, unsigned long duration, unsigned long steplimit)
{
  int i, s, s2, bc, l;
  unsigned long tea, t, t2, threshold, m, currentspd, lastspd, acc, samples;
  long load, lastload;
  
  float a;
  steps = 0;
  tea = millis();
  t = tea;

  driver.shaft(dir);

  bumped = false;

  l  = duration / 200 ;
  i = 1;
  currentspd = 0;
  m = t;

  while (!bumped) {

      a = abs(sin((M_PI ) * i / l));
      currentspd = start + int(spd * a);

 //     Log.printf("Speed %f %d %d (%d) steps %d\n", a, int(spd * a) ,currentspd, driver.cs2rms(driver.cs_actual()), steps);
      driver.VACTUAL(currentspd); //SET SPEED OF MOTOR

      t = m;

      while ((m - t) < 200)
      {
        m = millis();
        delay(10);
        if (bumped) break;
      }

      if (abs(m - tea) > duration) break;
      if (((i % 5) == 0) && islistening && get_motor_status) events.send(get_stats(),"motor",m);
      if (steps >= steplimit) break;
    i++;
  }
  driver.VACTUAL(0);
  
  t = millis() - tea;

  if (bumped) {
    Log.printf("bump %d\n", driver.SG_RESULT());
    digitalWrite(RED_LED, LOW);
    delay(200);
  }
  else if (steps >= steplimit)
    Log.printf("Max steps reached\n");
  else
    Log.printf("timed out\n");

  return t;
}


void MoveDoor(void *ptr) {
  unsigned long i;
  bool close = false;;

  while(true) {
    switch (direction)
    {
      case 0:
        driver.VACTUAL(0);
        
        if (get_motor_status)
        {
          get_motor_status = false;
          events.send(get_stats(),"motor",millis());
        }
        vTaskDelay(100);
        // Log.printf(".\n");
        digitalWrite(ENABLE_PIN ,HIGH);

        if (close)
        {
          if (autoclose_timer == 0)
          {
            autoclose_timer = millis();
          }
          else if ((millis() - autoclose_timer) > autoclose_delay)
          {
            direction = 1;
            autoclose = false;
            close = false;
          }
        }
        else
          direction = -2;
        break;
      case  1:
      case -1:

        digitalWrite(ENABLE_PIN ,LOW);
        Log.printf("-");
        shaft =  (direction == 1) ? 1 : 0; 
        digitalWrite(RED_LED, HIGH);
      
        Log.printf("Moving\n");

        i = move(12800, shaft ? 128 * 1500: 128 * 1500, shaft,shaft ? closetiming : opentiming, shaft ? stepstoclose : stepstoopen);
        Log.printf("Took %d msecs.\n", i);
        
        if (shaft)
          timetoclose = i;
        else
          timetoopen = i;

        digitalWrite(RED_LED, HIGH);
        vTaskDelay(1000);

        if (bumped) {
          Log.printf("backing up\n");
          i = move(12800, 0, !shaft, 2000, 200);
          Log.printf("Took %d msecs.\n", i);
          vTaskDelay(1000);
          digitalWrite(RED_LED, HIGH);
        }

        Log.printf("forward again\n");
        i = move(12800,  0, shaft, 8000, 200);
        Log.printf("Took %d msecs.\n", i);
        direction = 0;
        close = autoclose;
        autoclose_timer = 0;
        digitalWrite(RED_LED, HIGH);
        break;
    case -2:
      vTaskDelay(1000);
      break;
    }
  }
}

// Capture Photo and Save it to SPIFFS
void capturePhotoSaveSpiffs( void ) {
  camera_fb_t * fb = NULL; // pointer
  bool ok = 0; // Boolean indicating if the picture has been taken correctly
  unsigned long tea = millis();
  bool beentold = false;

    // Take a photo with the camera
    Log.printf("Taking a photo...\n");

  do {

    fb = esp_camera_fb_get();

    if (!fb) {
      Log.printf("Camera capture failed\n");
      return;
    }
    delay(100);

    if (fb->len > 1000)
    {
      delay(100);
      File file = SPIFFS.open(FILE_PHOTO, FILE_WRITE);

      // Insert the data in the photo file
      if (!file) {
        if (!beentold)
        {
          Log.printf("Failed to open file in writing mode\n");
          beentold = true;
        }
        ok = false;
      }
      else {
        file.write(fb->buf, fb->len); // payload (image), payload length
        file.close();
        file = SPIFFS.open(FILE_PHOTO);
        if (file.size() > 1000) {
          ok = true;
          Log.printf("The picture has been saved in %s  - Size: %d bytes\n", FILE_PHOTO, file.size());
        }
        file.close();
      }
    }
    esp_camera_fb_return(fb);
    if ((millis() - tea) > 10000) ok = true;
  } while ( !ok );
  digitalWrite(FLASH_PIN ,LOW);
}

void handleUpdate(AsyncWebServerRequest *request) {
  char* html = "<form method='POST' action='/doUpdate' enctype='multipart/form-data'><input type='file' name='update'><input type='submit' value='Update'></form>";
  request->send(200, "text/html", html);
}

void handleDoUpdate(AsyncWebServerRequest *request, const String& filename, size_t index, uint8_t *data, size_t len, bool final) {
  char b[100];
  static bool flip = false;

  static size_t content_len;
  static size_t saved_len;

  if (!index){
    Log.printf("Updating\n");
    content_len = request->contentLength();
    saved_len = 0;
    flip = false;
    // if filename includes spiffs, update the spiffs partition
    int cmd = U_FLASH;
    if (!Update.begin(UPDATE_SIZE_UNKNOWN, cmd)) {
      Update.printError(Log);
    }
  }
  
  if (Update.write(data, len) != len) {
    digitalWrite(RED_LED, LOW);
    Update.printError(Log);
  }
  else if (islistening) {
    saved_len += len;
    sprintf(b, "%d%%", 100 * saved_len / content_len);
    events.send(b ,"progress",millis());
    digitalWrite(RED_LED, flip ? HIGH : LOW);
    flip = !flip;
  }

  if (final) {
    digitalWrite(RED_LED, HIGH);
    AsyncWebServerResponse *response = request->beginResponse(302, "text/plain", "Please wait while the device reboots");
    response->addHeader("Refresh", "30");  
    response->addHeader("Location", "/");
    request->send(response);
    if (!Update.end(true)){
      Update.printError(Log);
    } else {
      Log.printf("Update complete\n");
      vTaskDelay(1000);
      ESP.restart();
    }
  }
}


